from pytest_smell.constants import CAST_KEYWORDS


def is_assertion(line):
    return 'assert' in line


def is_sleep(line):
    return line.startswith('sleep') or line.startswith('await asyncio.sleep')


def is_print(line):
    return line.startswith('print')


def is_annotation(line, annotation_name):
    return line.startswith(annotation_name)


def starts_try_block(line):
    return 'try:' in line


def starts_except_block(line):
    return 'except:' in line


def get_assertion_method(line, all_lines):
    # TODO: refactor this maybe using recursion
    asserted_entity = line[line.find(' '):line.find('==')].strip()
    if '(' not in asserted_entity:  # variable tested
        if '.' in asserted_entity:
            asserted_entity = asserted_entity.split('.')[0]
        for line in all_lines:
            if line.startswith(f'{asserted_entity} ='):
                return line[line.find(' ') + 2:line.find('(')].strip()
    else:
        first_call = line[line.find(' '):line.find('(')]  # From first space after assert until the function call
        if first_call in CAST_KEYWORDS:
            if '(' not in asserted_entity:  # cast variable tested
                if '.' in asserted_entity:
                    asserted_entity = asserted_entity.split('.')[0]
                for line in all_lines:
                    if line.startswith(f'{asserted_entity} ='):
                        return line[line.find(' ') + 2:line.find('(')].strip()
            return line[line.find('('):line.rfind(')')]  # cast method tested


def get_try_except_block(index_of_try_line, all_lines):
    try_except_block_content = []
    for line in all_lines[index_of_try_line:]:
        try_except_block_content.append(line)
        if starts_except_block(line):
            break
    return try_except_block_content


def contains_numeric_literals(assert_line):
    assertion_elements = assert_line.split('==')
    for el in assertion_elements:
        el = el.strip()
        try:
            int(el)
            float(el)
            return True
        except ValueError:
            pass
    return False



# TODO: Move this maybe to a universal config
# TODO: Add proper explanation messages
SMELL_INFO_MAPPING = {
    'duplicate_assert': {
        'display_name': 'Duplicate Assert',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test has %d duplicate assert(s):',
        'check_annotations': False
    },
    'eager_test': {
        'display_name': 'Eager Test',
        'metric_threshold': 1,
        'reversed_metric': False,
        'explanation_message': 'Test has %d methods tested:',
        'check_annotations': False
    },
    'conditional_logic': {
        'display_name': 'Conditional Logic',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test has %d control flow statement(s):',
        'check_annotations': False
    },
    'exception_handling': {
        'display_name': 'Exception Handling',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test is passed or failed upon %d method(s) explicitly throwing an exception:',
        'check_annotations': False
    },
    'sleepy_test': {
        'display_name': 'Sleepy Test',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test has %d sleep methods:',
        'check_annotations': False
    },
    'redundant_print': {
        'display_name': 'Redundant Print',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test has %d print methods:',
        'check_annotations': False
    },
    'ignored_test': {
        'display_name': 'Ignored Test',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test will not be executed due to the %d skip annotation:',
        'check_annotations': True
    },
    'unknown_test': {
        'display_name': 'Unknown Test',
        'metric_threshold': 1,
        'reversed_metric': True,
        'explanation_message': 'Test has to have at least 1 assertion statements, while it has %d',
        'check_annotations': True
    },
    'magic_number': {
        'display_name': 'Magic Number',
        'metric_threshold': 0,
        'reversed_metric': False,
        'explanation_message': 'Test has %d assertions which contain numeric literals',
        'check_annotations': False
    },
    'assertion_roulette': {
        'display_name': 'Assertion Roullete',
        'metric_threshold': 1,
        'reversed_metric': False,
        'explanation_message': 'Test has %d additional undocumented asserts:',
        'check_annotations': False
    },
}



import os
import re

from pytest_smell.utils import is_annotation


def traverse_tests_directories(root_path):
    for dir_path, _, files in os.walk(root_path):
        if '__pycache__' in dir_path:
            continue
        for filename in files:
            file_path = os.path.join(dir_path, filename)
            if (filename.startswith('test') and filename.endswith('.py')) or filename.endswith('test.py'):
                with open(file_path) as test_file:
                    yield file_path, test_file.readlines()


def traverse_tests_file(tests_directory, with_annotations=False):
    for file_path, lines in traverse_tests_directories(tests_directory):
        current_test_name, current_test_lines, current_test_annotations = None, [], []
        for line_number, line in enumerate(lines):
            line = line.strip('\n')
            if line.startswith('def test_'):
                current_test_name = re.search('def (.*)\\(', line).group(1)
            elif with_annotations and is_annotation(line, '@pytest'):
                current_test_annotations.append(line.strip())
            if current_test_name:
                if '    ' in line or line.startswith('def test_'):
                    current_test_lines.append(line.strip())
                    if line_number == len(lines) - 1:  # eof
                        test_lines = current_test_lines if not with_annotations else [*current_test_annotations,
                                                                                      *current_test_lines]
                        yield file_path, current_test_name, test_lines, line_number - len(current_test_lines)
                else:
                    next_two_lines_are_empty = not any([line, lines[line_number + 1].strip()])
                    if next_two_lines_are_empty:
                        test_lines = current_test_lines if not with_annotations else [*current_test_annotations,
                                                                                      *current_test_lines]
                        yield file_path, current_test_name, test_lines, line_number - len(current_test_lines)
                        current_test_name, current_test_lines, current_test_annotations = None, [], []
